/*
 * lockset.c
 *
 *  Created on: Oct 11, 2010
 *      Author: parkcs
 *
 *  Overview: Space and time efficient representation of locksets for Thrille
 *  Each lockset requires a multiple of 64 bits, depending on how many locks
 *  are used in the program. Uses coalesced hashing for efficient lookupg
 *
 *  Requirements: efficient set/clear, containment, intersection,
 *  and dump operations
 *
 *  Set (at upc_lock): During normal use, should be a simple bitwise OR (|) and
 *  adding the lock into the pool if not already in pool, but in special cases
 *  the lockset size (and the lockpool) may have to grow
 *
 *  Clear (at upc_unlock): simple bitwise AND (&) with 0 at correct position
 *
 *  Containment (at memory access): checks if superset == (superset | subset)
 *
 *  Intersection (at race detection): set1 & set2
 *
 *  Dump (at barrier boundaries): memory compact and consecutive
 *  representation of the pool and the bitsets
 */

#include <stdio.h>
#include "lockset.h"

void lock_print(const lock l) {
    printf(" 0x%"PRIx64, l);
}

int lock_comp(const void* x1, const void* x2) {
    lock l1 = *(const lock*) x1;
    lock l2 = *(const lock*) x2;
    // not using simple minus operator (i.e. l1 - l2) because of integer width issues
    return (l1 < l2) ? -1 : (l1 > l2 ? 1 : 0);
}

void lock_sorted_from_simple(memory_pool* tmp, lock_sorted* ls,
        const lock_pool_simple* p, const lock_set* s) {
    ls->length = lock_set_elems(s);
    ls->data = (lock*) allocate(tmp, sizeof (lock) * ls->length);

    int byteidx = 0;
    int pos = 0;
    do {
        int bitidx = 0;
        while (pos < ls->length && bitidx < POOL_SIZE) {
            char skip = NEXT_INDEX[(data_bits(s->bits[byteidx]) >> bitidx) & 255];
            bitidx += skip;
            if (skip < 8) {
                ls->data[pos++] = p->data[byteidx * POOL_SIZE + bitidx];
                ++bitidx;
            }
        }
    } while (more_bits(s->bits[byteidx++]));
}

void lock_sorted_print(const lock_sorted* ls) {
    printf("{");
    if (ls->length) {
        printf(" 0x%"PRIx64, ls->data[0]);
        for (int i = 1; i < ls->length; i++) {
            printf(", 0x%"PRIx64, ls->data[i]);
        }
    }
    printf(" }");
}
