/*
 * upct_racer.c -- online race detection for UPC
 *
 */

#ifndef UPCR_THRILLE_RACER
#define UPCR_THRILLE_RACER  // Make sure all recursive calls into thrille are properly suffixed
#endif

#include "upct_types.h"
#include "lockset/lockset.h"
#include "intervals/addr_interval_skiplist.h"
#include "race_trie.h"
#include "allocator.h"
#include "str_dbl_map.h"

//void add_myaccess(pivl mem, access_info*);
int other_access(addr_islist*, char*, int, int);
int shmem_access(addr_islist*, access_info_full*);
race_data_list* create_rd_list(memory_pool*, pivl_pool* p);
int dump_rd_list(race_data_list* rdl, char* buf);
void merge_islists(addr_islist* dst, addr_islist* src);


/* Shared Data */
shared int race_id;
shared [1] mem_access_db mem_access[THREADS];

/* Per Thread Data */
upc_lock_t * thrille_lock;
thrille_info_t thrille_info = {0, 0, 0, 0};
lock_pool* thrille_lockpool;
lock_set* thrille_lockset;
addr_islist* thrille_myaccess[THRILLE_PHASE_HISTORY];
addr_islist* thrille_otaccess[THREADS];
int my_activity = 0;
#ifndef NO_ASYNC_XFER
bupc_handle_t nbsend_handles[THRILLE_PHASE_HISTORY][THREADS];
#endif

// 0-3 for each past memory accesses on memory with my affinity, recycled after 4 phases
// 4   for accesses in this phase that will be shipped away, recycled at end of phase
// 5   for phase-independent stuff like lock_pool and lock_set, never recycled
memory_pool mempools[THRILLE_PHASE_HISTORY + 2];
#define current_phase_mempool (mempools + THRILLE_PHASE_HISTORY)
#define PERSISTENT_MEM	(THRILLE_PHASE_HISTORY + 1)
#define persistent_mempool (mempools + PERSISTENT_MEM)

/* statistic counters */
upct_stat mystats;
extern bupc_tick_t comm_stat, search_stat, oa_stat;
bupc_tick_t comm_cur, search_cur, oa_cur;

str_dbl_map* prob_map;

int rand_test(const char* sig) {
    double prob = str_dbl_map_get(prob_map, sig, 1.0);
    if (prob > EPS && rand() < (prob * RAND_MAX)) {
        str_dbl_map_put(prob_map, sig, prob * BACK_OFF);
        return TRUE;
    }
    return FALSE;
}

void thrille_init(void) {
    int myid = MYTHREAD;
    if (myid == 0) {
        fprintf(stderr, "Initializing Race Detection " THRILLE_VERSION " (BACK_OFF=%g, EPS=%g)...\n", BACK_OFF, EPS);
        race_id = 0;
    }

    /* Global Allocations */
    thrille_lock = upc_all_lock_alloc();

    new_mempool(persistent_mempool);
    /* Local Allocations */
    thrille_info.inside_inst = 0;
    thrille_info.phase = 0;
    thrille_lockpool = lock_pool_new(persistent_mempool);
    thrille_lockset = lock_set_new(persistent_mempool);

    for (int i = 0; i < THRILLE_PHASE_HISTORY; i++) {
        new_mempool(mempools + i);
        thrille_myaccess[i] = new_addr_islist(mempools + i);
        mem_access[myid].pp[i].activity = 0;
        for (int j = 0; j < THREADS; j++) {
#ifndef NO_ASYNC_XFER
            nbsend_handles[i][j] = BUPC_COMPLETE_HANDLE;
#endif
            *((int*)mem_access[myid].pp[i].in[j]) = 0;
        }
    }

    new_mempool(current_phase_mempool);

    /*thrille_otaccess = (addr_islist**) allocate(persistent_mempool,
            sizeof (void*) *THREADS);*/
    for (int i = 0; i < THREADS; i++) {
        thrille_otaccess[i] = NULL;
    }

    prob_map = str_dbl_map_new(persistent_mempool, 1023);
    
#ifdef THRILLE_STATS
    open_stat(&mystats, MYTHREAD);
    comm_cur = comm_stat = search_cur = search_stat = oa_cur = oa_stat = 0;
#endif

    upc_barrier;

    thrille_info.initialized = 1;
}

void thrille_before_read_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#if THRILLE_DEBUG & THRILLE_DEBUG_ACCESS
    fprintf(stderr, "[%d] Before Read of Memory 0x%"PRIx64"-0x%"PRIx64"\n", myid, mem_addr, mem_addr + nBytes);
#endif

#ifdef THRILLE_STATS
    search_cur = bupc_ticks_now();
#endif

    if (/*myid == owner ||*/ rand_test(sig)) {
        access_info_full acc;
        acc.mem.start = mem_addr;
        acc.mem.end = mem_addr + nBytes;
        //acc.mem.type = RIGHT_OPEN;

        acc.ai.thread = myid;
        acc.ai.access = READ;
        acc.ai.sig = sig;
        acc.ai.phase = thrille_info.phase;

        ++mystats.n_shread;

        lock_sorted_new(current_phase_mempool, &acc.ls, thrille_lockpool,
                thrille_lockset);

        // Add race data to corresponding trie based on owner
        int added;
        if (myid == owner)
            added = shmem_access(thrille_myaccess[thrille_phase_id], &acc);
        else {
            if (!thrille_otaccess[owner])
                thrille_otaccess[owner] = new_addr_islist(current_phase_mempool);
            added = shmem_access(thrille_otaccess[owner], &acc);
        }
        if (!added)
            ++mystats.n_strongerread;
        else
            my_activity = 1;

        lock_sorted_free(&acc.ls);
    }

#ifdef THRILLE_STATS
    search_stat += bupc_ticks_now() - search_cur;
#endif

    THRILLE_END_FUNC;
}

void thrille_after_read_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

uintmax_t thrille_after_read_reg_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        uintmax_t retval,
        int nBytes,
        int owner) {
    thrille_after_read_racer(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

float thrille_after_read_float_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        float retval,
        int nBytes,
        int owner) {
    thrille_after_read_racer(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

double thrille_after_read_double_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        double retval,
        int nBytes,
        int owner) {
    thrille_after_read_racer(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

void thrille_before_write_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#if THRILLE_DEBUG & THRILLE_DEBUG_ACCESS
    fprintf(stderr, "[%d] Before Write of Memory 0x%"PRIx64"-0x%"PRIx64"\n", myid, mem_addr, mem_addr + nBytes);
#endif
    
#ifdef THRILLE_STATS
    search_cur = bupc_ticks_now();
#endif

    if (/*myid == owner ||*/ rand_test(sig)) {
        access_info_full acc;
        acc.mem.start = mem_addr;
        acc.mem.end = mem_addr + nBytes;
        //acc.mem.type = RIGHT_OPEN;

        acc.ai.thread = myid;
        acc.ai.access = WRITE;
        acc.ai.sig = sig;
        acc.ai.phase = thrille_info.phase;

        lock_sorted_new(current_phase_mempool, &acc.ls, thrille_lockpool,
                thrille_lockset);

        ++mystats.n_shwrite;
        // Add race data to corresponding trie based on owner
        int added;
        if (myid == owner)
            added = shmem_access(thrille_myaccess[thrille_phase_id], &acc);
        else {
            if (!thrille_otaccess[owner])
                thrille_otaccess[owner] = new_addr_islist(current_phase_mempool);
            added = shmem_access(thrille_otaccess[owner], &acc);
        }
        if (!added)
            ++mystats.n_strongerwrite;
        else
            my_activity = 1;


        lock_sorted_free(&acc.ls);
    }

#ifdef THRILLE_STATS
    search_stat += bupc_ticks_now() - search_cur;
#endif

    THRILLE_END_FUNC;
}

void thrille_after_write_racer(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

void thrille_before_lock_racer(int myid,
        uint64_t lock,
        const char* sig) {
}

void thrille_after_lock_racer(int myid,
        uint64_t lock,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //THRILLE_LOCK;

#if THRILLE_DEBUG & THRILLE_DEBUG_LOCK
    fprintf(stderr, "[%d] After lock 0x%"PRIx64"\n", myid, lock);
#endif
    thrille_lockset = lock_set_add(persistent_mempool, thrille_lockpool,
            thrille_lockset, lock);

    //THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_before_lock_attempt_racer(int myid,
        uint64_t lock,
        const char* sig) {
}

int thrille_after_lock_attempt_racer(int myid,
        uint64_t lock,
        const char* sig,
        int retval) {

#if THRILLE_DEBUG & THRILLE_DEBUG_LOCK
    fprintf(stderr, "[%d] After lock_attempt 0x%"PRIx64" (%s)\n", myid, lock, retval ? "success" : "fail");
#endif

    if (retval) {
        thrille_after_lock_racer(myid, lock, sig);
    }

    return retval;
}

void thrille_before_unlock_racer(int myid,
        uint64_t lock,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //THRILLE_LOCK;

#if THRILLE_DEBUG & THRILLE_DEBUG_LOCK
    fprintf(stderr, "[%d] Before unlock 0x%"PRIx64"\n", myid, lock);
#endif
    lock_set_remove(thrille_lockpool, thrille_lockset, lock);

    //THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_after_unlock_racer(int myid,
        uint64_t lock,
        const char* sig) {
}

/* Returns number of bytes initiated for sending */
int dump_all_rd_list(int myid) {
    int totsize = 0;
#ifdef THRILLE_STATS
    comm_cur = bupc_ticks_now();
#endif
    if (my_activity) {
        int phase = thrille_phase_id;
        for (int t = 0; t < THREADS; ++t) {
            if (thrille_otaccess[t] && t != myid) {
                char* buf = (char*)mem_access[myid].pp[phase].out[t];
                race_data_list* rdl = create_rd_list(
                        current_phase_mempool,
                        thrille_otaccess[t]->interval_pool);
                int size = dump_rd_list(rdl, buf);
                if (size > 0) {
#ifndef NO_ASYNC_XFER
                    nbsend_handles[phase][t] = bupc_memput_async(
                            mem_access[t].pp[phase].in[myid], buf, size);
                    bupc_memset_asynci(&mem_access[t].pp[phase].activity, 1, 1);
#else
                    upc_memput(mem_access[t].pp[phase].in[myid], buf, size);
                    mem_access[t].pp[phase].activity = 1;
#endif
#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
                    fprintf(stderr, "[%d] Dumped %d bytes to thread %d phase %d\n",
                        myid, size, t, thrille_info.phase);
#endif
                    totsize += size;
                }
#ifndef NO_ASYNC_XFER
                else {
                    nbsend_handles[phase][t] = BUPC_COMPLETE_HANDLE;
                }
#endif
                thrille_otaccess[t] = NULL;
            }
        }
        my_activity = 0;
    }
#ifdef THRILLE_STATS
    comm_stat += bupc_ticks_now() - comm_cur;
#endif
    return totsize;
}

void thrille_before_notify_racer(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    //THRILLE_LOCK;
#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] Before notify at phase %d\n", myid, thrille_info.phase);
#endif
    int pm2 = thrille_phase_idn(thrille_info.phase - 2);
    int pm1 = thrille_phase_idn(thrille_info.phase - 1);
#ifdef THRILLE_STATS
    comm_cur = bupc_ticks_now();
#endif
#ifndef NO_ASYNC_XFER
    bupc_waitsynci();
    bupc_waitsync_all (nbsend_handles[pm2], THREADS);
    bupc_waitsync_all (nbsend_handles[pm1], THREADS);
#endif
#ifdef THRILLE_STATS
    comm_stat += bupc_ticks_now() - comm_cur;
#endif
    mystats.n_bytesent = dump_all_rd_list(myid);

#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
    pivl_pool_print(thrille_myaccess[thrille_phase_id]->interval_pool);
    //addr_islist_print(thrille_myaccess[thrille_phase_id]);
#endif
    //THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_after_notify_racer(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
#ifdef THRILLE_STATS
    mystats.sz_istree = pivl_pool_elems(thrille_myaccess[thrille_phase_id]->interval_pool);
    write_stat(&mystats, thrille_info.phase);
#endif
    thrille_info.phase++;
    clear_stat(mystats);

#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] Now in phase %d (After notify)\n", myid, thrille_info.phase);
#endif

#ifdef THRILLE_STATS
    oa_cur = bupc_ticks_now();
#endif
    //pivl_pool_free(thrille_myaccess[thrille_phase_id]->interval_pool);
    recycle_mempool(mempools + thrille_phase_id);
    thrille_myaccess[thrille_phase_id] = new_addr_islist(mempools + thrille_phase_id);
    /*for (int i = 0; i < THREADS; i++) {
        pivl_pool_free(thrille_otaccess[i]->interval_pool);
    }*/
    recycle_mempool(current_phase_mempool);
    /*for (int i = 0; i < THREADS; i++) {
        thrille_otaccess[i] = new_addr_islist(current_phase_mempool);
    }*/
#ifdef THRILLE_STATS
    oa_stat += bupc_ticks_now() - oa_cur;
#endif

    THRILLE_END_FUNC;
}

void thrille_before_wait_racer(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    //THRILLE_LOCK;
#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] Before wait at phase %d\n", myid, thrille_info.phase);
#endif

#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
    pivl_pool_print(thrille_myaccess[thrille_phase_id]->interval_pool);
    //addr_islist_print(thrille_myaccess[thrille_phase_id]);
#endif
    //THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

int read_all_rd_list(int myid, int srcp) {
    int totsize = 0;
#ifdef THRILLE_STATS
    oa_cur = bupc_ticks_now();
#endif
    int srcphidx = thrille_phase_idn(srcp);
    //int dstphidx = thrille_phase_idn(thrille_info.phase - 2);

    if (mem_access[myid].pp[srcphidx].activity) {
        for (int t = (myid + 1) % THREADS; t != myid; t = (t + 1) % THREADS) {
            char* in = (char*) mem_access[myid].pp[srcphidx].in[t];
            int size = *((int*)in);
            if (size) {
                totsize += other_access(thrille_myaccess[srcphidx], in, srcp, t);
                *((int*)in) = 0;
            }
        }
        mem_access[myid].pp[srcphidx].activity = 0; 
    }
#ifdef THRILLE_STATS
    oa_stat += bupc_ticks_now() - oa_cur;
#endif
    return totsize;
}

void thrille_after_wait_racer(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();

    // initiate transfer to other threads at [phase]
    mystats.n_bytesent = dump_all_rd_list(myid);

    // add dumped data from other threads at [phase-2]
    mystats.n_byterecv = read_all_rd_list(myid, thrille_info.phase - 2);

#ifdef THRILLE_STATS
    search_cur = bupc_ticks_now();
#endif
    int pm3 = thrille_phase_idn(thrille_info.phase - 3);
    int pm2 = thrille_phase_idn(thrille_info.phase - 2);
    int pm1 = thrille_phase_idn(thrille_info.phase - 1);
    // check for races between pp[phase-2] and pp[phase-3]
    merge_islists(thrille_myaccess[pm3], thrille_myaccess[pm2]);
    // merge my accesses from pp[phase-1] into pp[phase-2]
    merge_islists(thrille_myaccess[pm1], thrille_myaccess[pm2]);
#ifdef THRILLE_STATS
    search_stat += bupc_ticks_now() - search_cur;
#endif

    // merge other accesses from pp[phase-1] into pp[phase-2]
    mystats.n_byterecv += read_all_rd_list(myid, thrille_info.phase - 1);

    // make pp[phase-1] point to pp[phase-2]
    //pivl_pool_free(thrille_myaccess[pm1]->interval_pool);
    //thrille_myaccess[pm1] = thrille_myaccess[pm2];

#ifdef THRILLE_STATS
    mystats.sz_istree = pivl_pool_elems(thrille_myaccess[thrille_phase_id]->interval_pool);
    write_stat(&mystats, thrille_info.phase);
#endif
    //str_dbl_map_print(prob_map);
    thrille_info.phase++;
    clear_stat(mystats);

#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] Now in phase %d (After wait)\n", myid, thrille_info.phase);
#endif

#ifdef THRILLE_STATS
    oa_cur = bupc_ticks_now();
#endif
    //pivl_pool_free(thrille_myaccess[thrille_phase_id]->interval_pool);
    recycle_mempool(mempools + thrille_phase_id);
    thrille_myaccess[thrille_phase_id] = new_addr_islist(mempools + thrille_phase_id);
    /*for (int i = 0; i < THREADS; i++) {
        pivl_pool_free(thrille_otaccess[i]->interval_pool);
    }*/
    recycle_mempool(current_phase_mempool);
    /*for (int i = 0; i < THREADS; i++) {
        thrille_otaccess[i] = new_addr_islist(current_phase_mempool);
    }*/
#ifdef THRILLE_STATS
    oa_stat += bupc_ticks_now() - oa_cur;
#endif
    //THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_before_try_wait_racer(int myid, int barrier_val, const char* sig) {
    // Rewrite data only if new data to transfer <= no longer necessary
    // since data transfers are done after_wait
#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] Before try_wait at phase %d\n", myid, thrille_info.phase);
#endif
    thrille_before_wait_racer(myid, barrier_val, sig);
}

int thrille_after_try_wait_racer(int myid, int barrier_val, const char* sig, int retval) {
#if THRILLE_DEBUG & THRILLE_DEBUG_BARRIER
    fprintf(stderr, "[%d] After try_wait (%s)\n", myid, retval ? "success" : "fail");
#endif

    if (retval) {
        thrille_after_wait_racer(myid, barrier_val, sig);
    }
    return retval;
}

int other_access(addr_islist* isl, char* dump, int phase, int thread) {
    char* odump = dump;
    lock_pool_simple lp;
    int lpbytes = *((int*) dump);
    lp.length = (lpbytes - sizeof (int)) / sizeof (lock);
    lp.data = (lock*) (dump + sizeof (int));
    dump += lpbytes;
    while (*((int*) dump)) {
        race_data* rd = (race_data*) dump;

#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
        print_racedata(rd);
#endif

        access_info_full acc;
        acc.ai.thread = thread;
        acc.ai.access = rd->access;
        acc.ai.sig = rd->sig;
        acc.ai.phase = phase;

        acc.mem.start = rd->addr.start;
        acc.mem.end = rd->addr.end;
        //acc.mem.type = RIGHT_OPEN;

        lock_sorted_from_simple(current_phase_mempool, &acc.ls, &lp, &(rd->ls));
        shmem_access(isl, &acc);

        //lock_sorted_free(&acc.ls);

        int ls_size = lock_set_size(&rd->ls);
        int rd_size = sizeof (race_data) + sizeof (lock_set)*(ls_size - 1);
        dump += rd_size;
    }
    return (int)(dump - odump);
}

int shmem_access(addr_islist* isl, access_info_full* acc) {
#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
    fprintf(stderr, "[%d] Processing access: ", MYTHREAD);
    print_access(&acc->mem, &acc->ai, MYTHREAD);
#endif

    // get overlap
    pivl_set* overlap = addr_islist_overlap_interval(current_phase_mempool,
            isl, &acc->mem);

    // Check for weaker accesses in overlapping tries
    pivl_iter it;
    pivl_iter_init(&it, isl->interval_pool, overlap);
    while (pivl_iter_hasnext(&it)) {
        pivl other = *pivl_iter_next(&it);
        if (contains_interval(other->range, acc->mem)) {
            access_info* otherai;
            if ((otherai = race_trie_check_weaker(other->trie, acc))) {
#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
                fprintf(stderr, "[%d] Weaker access found: ", MYTHREAD); //%s ", otherai->sig);
                print_access(&other->range, otherai, MYTHREAD);
                //pivl_print(other);
                //fprintf(stderr, "(%s)\n", otherai->access == READ ? "Read" : "Write");
#endif
                pivl_set_free(overlap);
                return 0;
            }
        }
    }

    // check for racing access in overlapping tries
    pivl_iter_init(&it, isl->interval_pool, overlap);
    while (pivl_iter_hasnext(&it)) {
        pivl other = *pivl_iter_next(&it);
        access_info* otherai;
        if ((otherai = race_trie_check_race(other->trie, acc))) {
            int id;
            THRILLE_LOCK;
            id = ++race_id;
            THRILLE_UNLOCK;
            fprintf(stderr, "[%d] Potential race #%d found:\n", MYTHREAD, id);
            print_access(&acc->mem, &acc->ai, MYTHREAD);
            print_access(&other->range, otherai, MYTHREAD);
            write_race_pair(id, MYTHREAD, &acc->mem, &acc->ai, &other->range, otherai);
            
            str_dbl_map_put(prob_map, acc->ai.sig, 0.0);
            str_dbl_map_put(prob_map, otherai->sig, 0.0);
            // TODO: also set prob. for other->sig to 0.0?
            break;
        }
    }


    // add to trie and remove stronger
    memory_pool* mp = isl->endpoints->mempool;
    BOOL inserted = FALSE;
    pivl_iter_init(&it, isl->interval_pool, overlap);
    while (pivl_iter_hasnext(&it)) {
        // remove stronger access in overlapping tries
        pivl other = *pivl_iter_next(&it);
        if (contains_interval(acc->mem, other->range)) {
            race_trie_remove_stronger(other->trie, acc);
            if (eq_interval(acc->mem, other->range)) {
                race_trie_merge(mp, other->trie, acc);
                inserted = TRUE;
            }
        }

    }
    if (!inserted) {
        pivl mem = (pivl) allocate(mp, sizeof (ivl));
        mem->range = acc->mem;
        mem->trie = race_trie_create(mp, &acc->ls, &acc->ai);
        addr_islist_insert(current_phase_mempool, isl, mem);
    }

    pivl_set_free(overlap);
    return 1;
}

race_data_list* add_rd_list(memory_pool* mp,
        addr_interval addr, race_trie_node* rt, race_data_list* rdl, lock_set* ls) {
    if (rt->data) {
        int ls_size = lock_set_size(ls);
        int rd_size = sizeof (race_data) + sizeof (lock_set)*(ls_size - 1);
        race_data* rd = (race_data*) allocate(mp, rd_size);

        rd->addr.start = addr.start;
        rd->addr.end = addr.end;
        rd->access = rt->data->access;
        rd->sig = rt->data->sig;
        memcpy(&(rd->ls), ls, sizeof (lock_set) * ls_size);

#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
        print_racedata(rd);
#endif

        race_data_list* newhead = (race_data_list*) allocate(mp, sizeof (race_data_list));
        newhead->rd = rd;
        newhead->rd_size = rd_size;
        newhead->next = rdl;
        rdl = newhead;
    }
    // traverse children
    trie_ch_node* c = rt->children;
    while (c) {
        ls = lock_set_add(mp, thrille_lockpool, ls, c->lock);
        rdl = add_rd_list(mp, addr, c->node, rdl, ls);
        lock_set_remove(thrille_lockpool, ls, c->lock);
        c = c->next;
    }

    return rdl;
}

race_data_list* create_rd_list(memory_pool* mp, pivl_pool* p) {
    race_data_list* rdl = 0;
    while (p) {
        for (int i = 0; i < p->filled; i++) {
            lock_set* ls = lock_set_new(mp);
            rdl = add_rd_list(mp, p->data[i]->range, p->data[i]->trie, rdl, ls);
            lock_set_free(ls);
        }
        p = p->next;
    }
    return rdl;
}

/* rd_list dump structure:
   | lp_bytes | lp_data... | race_data | race_data | ... | 0 |
   Note: last 0 is to protect read of list beyond tail
 */
int dump_rd_list(race_data_list* rdl, char* op) {
    char* p = op;
    race_data_list* head = rdl;
    int rd_size = 0, tot_size = 0;
    while (head) {
        rd_size += head->rd_size;
        head = head->next;
    }

    if (rd_size) {
        int lp_size = sizeof (int) + 
                sizeof (lock) * lock_pool_elems(thrille_lockpool);
        //		tot_size = rd_size + lp_size + sizeof(int);

        //		mem_access[myt].pp[thrille_phase_id].size[t] = tot_size;

#if 0
        if (tot_size > MEM_ACC_BLOCK_SIZE)
            out_of_memory(tot_size);
#endif
        //upc_free(mem_access[myt].pp[thrille_phase_id].data[t]);
        //char* p = (char*)mem_access[myt].pp[thrille_phase_id].data[t];


        *((int*) p) = lp_size;
        p += sizeof (int);

        lock_pool* lp = thrille_lockpool;
        while (lp) {
            int size = sizeof (lock) * lp->filled;
            memcpy(p, lp->data, size);
            p += size;
            lp = lp->next;
        }

        head = rdl;
        while (head) {
            int size = head->rd_size;
            memcpy(p, head->rd, size);
            p += size;
            head = head->next;
        }

        *((int*) p) = 0;
        p += sizeof (int);
    }
    /*    else {
            mem_access[myt].pp[thrille_phase_id].size[t] = 0;
        }
     */

#if THRILLE_DEBUG & THRILLE_DEBUG_VERBOSE
    fprintf(stderr, "[%d] Dumped %d bytes for thread ? to [%"PRIx64",%"PRIx64")\n",
            MYTHREAD, (int) (p - op), (uint64_t) op, (uint64_t) p);
    /*    fprintf(stderr, "[%d] Dumped %d bytes for thread %d to [%"PRIx64",%"PRIx64") %d\n",
                MYTHREAD, tot_size, t, (uint64_t)mem_access[myt].pp[thrille_phase_id].data[t],
                (uint64_t)p, (int)(p - mem_access[myt].pp[thrille_phase_id].data[t]));*/
#endif
    return p-op;
}

void _merge_islists_rec(addr_islist* dst, addr_interval addr,
        race_trie_node* rt, lock_set* ls) {
    if (rt->data) {
        access_info_full acc;
        memcpy(&acc.ai, rt->data, sizeof (access_info));
        memcpy(&acc.mem, &addr, sizeof (addr_interval));
        lock_sorted_new(current_phase_mempool, &acc.ls, thrille_lockpool, ls);

        shmem_access(dst, &acc);

        lock_sorted_free(&acc.ls);
    }
    // traverse children
    trie_ch_node* c = rt->children;
    while (c) {
        ls = lock_set_add(current_phase_mempool, thrille_lockpool,
                ls, c->lock);
        _merge_islists_rec(dst, addr, c->node, ls);
        lock_set_remove(thrille_lockpool, ls, c->lock);
        c = c->next;
    }

}

void merge_islists(addr_islist* dst, addr_islist* src) {
    pivl_pool* p = src->interval_pool;
    while (p) {
        for (int i = 0; i < p->filled; i++) {
            lock_set* ls = lock_set_new(current_phase_mempool);
            _merge_islists_rec(dst, p->data[i]->range,
                    p->data[i]->trie, ls);
            lock_set_free(ls);
        }
        p = p->next;
    }
}

#include "upct_common.c"
