/*
 * upc_tester.c -- active testing for upc
 *
 */

#ifndef UPCR_THRILLE_TESTER
#define UPCR_THRILLE_TESTER  // Make sure all recursive calls into thrille are properly suffixed
#endif
#include "upct_types.h"

#include "upct_common.c"

#define UPCT_RACE_ID_ENV_STRING	"UPCT_RACE_ID"
#define UPCT_RACE_RESOLVE_ENV_STRING	"UPCT_RACE_RESOLVE_DIR"

struct _race_target_t {
    const char* sig;
};


/* Shared Data */
typedef bupc_sem_t* ptr_sem;
shared ptr_sem *semaphores;
shared race_info_t *races;
shared int num_thrash = 0;
shared int num_reproduced = 0;

shared double gprob = 1.0;


/* Per Thread Data */
upc_lock_t *thrille_lock;
thrille_info_t thrille_info;
race_info_t target1, target2;
int race_resolve_dir;
double lprob = 1.0;

void thrille_init(void) {
    if (MYTHREAD == 0) {
        fprintf(stderr, "UPC Active Testing " THRILLE_VERSION " started (BACK_OFF=%g, EPS=%g)...\n", BACK_OFF, EPS);
    }

    char* race_id_str = bupc_getenv(UPCT_RACE_ID_ENV_STRING);
    int id = race_id_str ? atoi(race_id_str) : 1;
    char* race_resolve_str = bupc_getenv(UPCT_RACE_RESOLVE_ENV_STRING);
    race_resolve_dir = race_resolve_str ? atoi(race_resolve_str) : 0;
    if (MYTHREAD == 0)
        fprintf(stderr, "Reading race pair %d\n", id);

    read_race_pair(&target1, &target2, id);

    if (MYTHREAD == 0) {
        print_race_info_ns(&target1);
        fprintf(stderr, "\n===\n");
        print_race_info_ns(&target2);
        fprintf(stderr, "\n");
        fprintf(stderr, "Resolve order: %d\n", race_resolve_dir);
    }

    //setRaceTargets();

    thrille_lock = upc_all_lock_alloc();
    semaphores = (shared ptr_sem*) upc_all_alloc(THREADS, sizeof (ptr_sem));
    semaphores[MYTHREAD] = bupc_sem_alloc(0);

    races = (shared race_info_t*) upc_all_alloc(THREADS, sizeof (race_info_t));
    races[MYTHREAD].addr[0] = 0;
    races[MYTHREAD].addr[1] = 0;

    thrille_info.inside_inst = 0;
    thrille_info.phase = 0;

    upc_barrier;

    thrille_info.initialized = 1;
}

int isRacingPair(const char* sig, int owned) {
    //fprintf(stderr, "%s %s\n", sig, target1.sig);
    int isRacingPair1 = (strcmp(sig, target1.sig) == 0)
            && (owned == target1.owned)
            //&& (thrille_info.phase == target1.phase)
            ;

    int isRacingPair2 = (strcmp(sig, target2.sig) == 0)
            && (owned == target2.owned)
            //&& (thrille_info.phase == target2.phase)
            ;

    return isRacingPair1 + 2 * isRacingPair2;
}

int checkPending(race_info_t event) {
    for (int racingThread = 0; racingThread < THREADS; racingThread++) {
        if (racingThread == event.thread) {
            continue;
        }
        race_info_t other = races[racingThread];
        int sameAddr = (other.addr[0] < event.addr[1] && event.addr[0] < other.addr[1]);
        int oneWrite = (event.access == WRITE ||
                other.access == WRITE);
        int sameTarget = (event.race_reported == 1 && other.race_reported == 1)
                || (event.race_reported == 2 && other.race_reported == 2);
        if (sameAddr) {
            if (oneWrite && !sameTarget)
                //if (sameAddr && oneWrite && !sameTarget)
                return racingThread;
            else
                return BAD_RACE;
        } else if (sameTarget)
            return BAD_RACE;
    }
    return NO_RACE;
}

void raceFound(race_info_t ev1, shared race_info_t* ev2) {
    fprintf(stderr, "REAL RACE CONFIRMED (p=%lf) - Racing pairs:\n", lprob);
    fprintf(stderr, "\t(T%d:0x%"PRIx64"-%"PRIx64":%s:%s)\n", ev1.thread, ev1.addr[0], ev1.addr[1],
            ev1.access == READ ? "R" : "W", ev1.sig);
    fprintf(stderr, "\t(T%d:0x%"PRIx64"-%"PRIx64":%s:%s)\n", ev2->thread, ev2->addr[0], ev2->addr[1],
            ev2->access == READ ? "R" : "W", ev2->sig);
}

void registerWaiting(race_info_t event) {
    races[MYTHREAD] = event;
}

void unregisterWaiting(race_info_t event) {
    races[MYTHREAD].addr[0] = 0;
    races[MYTHREAD].addr[1] = 0;
}

void printRaceInfoT(race_info_t event) {
    fprintf(stderr, "race_info_t(");
    fprintf(stderr, "Thread: %d, ", event.thread);
    fprintf(stderr, "Addr: 0x%"PRIx64"-%"PRIx64", ", event.addr[0], event.addr[1]);
    fprintf(stderr, "Access Type: %s, ", event.access == WRITE ? "W" : "R");
    fprintf(stderr, "Sig: %s, ", event.sig);
    fprintf(stderr, ")\n");
}

int randTest() {
    return rand() < (lprob * RAND_MAX);
}

void randomActive(race_info_t event) {
    THRILLE_LOCK;
    int racingThread = checkPending(event);
    if (racingThread == NO_RACE) {
        if (randTest()) {
            registerWaiting(event);
            int n = 0;
            fprintf(stderr, "[%d] @p%d paused at %s to try to create a real race (Addr: 0x%"PRIx64"-%"PRIx64"%s)\n",
                    MYTHREAD, thrille_info.phase, event.sig, event.addr[0], event.addr[1],
                    event.access == WRITE ? "W" : "R");
            int realRace = 0;
            while (!(realRace = bupc_sem_try(semaphores[MYTHREAD])) &&
                    n++ < MAX_SEM_TRIES) {
                THRILLE_UNLOCK;
                usleep(SEM_WAIT_USEC);
                THRILLE_LOCK;
            }
            fprintf(stderr, "[%d] @p%d is no longer paused\n", MYTHREAD, thrille_info.phase);
            unregisterWaiting(event);
            if (realRace) {
                //++num_reproduced;
                //fprintf(stderr, "%d reproduced (%d thrashes)\n", num_reproduced, num_thrash);
                if ((race_resolve_dir && event.race_reported == 1) ||
                        (!race_resolve_dir && event.race_reported == 2)) {
                    THRILLE_UNLOCK;
                    //fprintf(stderr, "[%d] Waiting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
                    usleep(10 * SEM_WAIT_USEC);
                    THRILLE_LOCK;
                }
                lprob = 0.0;
            } else {
                //++num_thrash;
                //fprintf(stderr, "%d thrashes (%d reproduced)\n", num_thrash, num_reproduced);
                lprob *= BACK_OFF;
            }
        }
    } else if (racingThread != BAD_RACE) {
        raceFound(event, &races[racingThread]);
        lprob = gprob = 0.0;
        bupc_sem_post(semaphores[racingThread]);
        if ((race_resolve_dir && event.race_reported == 1) ||
                (!race_resolve_dir && event.race_reported == 2)) {
            THRILLE_UNLOCK;
            //fprintf(stderr, "[%d] Posting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
            usleep(10 * SEM_WAIT_USEC);
            THRILLE_LOCK;
        }
    }
    THRILLE_UNLOCK;
}

void thrille_before_read_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] Before Read of Memory 0x%"PRIx64"-0x%"PRIx64" at %s\n", myid, mem_addr, mem_addr + nBytes, sig);
#endif

    int r;
    int owned = (MYTHREAD == owner) ? 1 : 0;
    if ((lprob > EPS) && (r = isRacingPair(sig, owned))) {
        race_info_t raceEvent;
        raceEvent.addr[0] = mem_addr;
        raceEvent.addr[1] = mem_addr + nBytes;
        raceEvent.thread = myid;
        raceEvent.access = READ;
        raceEvent.sig = sig;
        //ls_strcpy(file, raceEvent.file);
        raceEvent.phase = thrille_info.phase;
        //ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(raceEvent);
    }
    THRILLE_END_FUNC;
}

void thrille_after_read_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

uintmax_t thrille_after_read_reg_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        uintmax_t retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

float thrille_after_read_float_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        float retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

double thrille_after_read_double_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        double retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

void thrille_before_write_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] Before Write of Memory 0x%"PRIx64"-0x%"PRIx64" at %s\n", myid, mem_addr, mem_addr + nBytes, sig);
#endif

    int r;
    int owned = (MYTHREAD == owner) ? 1 : 0;
    if ((lprob > EPS) && (r = isRacingPair(sig, owned))) {
        race_info_t raceEvent;
        raceEvent.thread = myid;
        raceEvent.addr[0] = mem_addr;
        raceEvent.addr[1] = mem_addr + nBytes;
        raceEvent.access = WRITE;
        raceEvent.sig = sig;
        //ls_strcpy(file, raceEvent.file);
        raceEvent.phase = thrille_info.phase;
        //ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(raceEvent);
    }
    THRILLE_END_FUNC;
}

void thrille_after_write_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

void thrille_before_lock_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

void thrille_after_lock_tester(int myid,
        uint64_t lock_addr,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //fprintf(stderr, "[%d] After lock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_before_lock_attempt_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

int thrille_after_lock_attempt_tester(int myid,
        uint64_t lock,
        const char* sig,
        int retval) {

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] After lock_attempt 0x%"PRIx64" (%s)\n", myid, lock, retval ? "success" : "fail");
#endif

    if (retval) {
        thrille_after_lock_tester(myid, lock, sig);
    }

    return retval;
}

void thrille_before_unlock_tester(int myid,
        uint64_t lock_addr,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //fprintf(stderr, "[%d] Before unlock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_after_unlock_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

void thrille_before_notify_tester(int myid, int barrier_val, const char* sig) {
}

void thrille_after_notify_tester(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //fprintf(stderr, "[%d] Now in phase %d (After notify)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}

void thrille_before_wait_tester(int myid, int barrier_val, const char* sig) {
}

void thrille_after_wait_tester(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //fprintf(stderr, "[%d] Now in phase %d (After wait)\n", myid, thrille_info.phase);
    if (lprob >= EPS) {
        lprob *= gprob;
    }
    THRILLE_END_FUNC;
}

void thrille_before_try_wait_tester(int myid, int barrier_val, const char* sig) {
}

int thrille_after_try_wait_tester(int myid, int barrier_val, const char* sig, int retval) {
    if (retval) {
        thrille_after_wait_tester(myid, barrier_val, sig);
    }
    return retval;
}

